
# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '1.6.0'
version = '1.6.0'
full_version = '1.6.0'
git_revision = 'ab1c0907fe9255582397db04592d6066745018d3'
release = True

if not release:
    version = full_version
